@@ -34,15 +34,23 @@
 
 #include "list.h"
 #include "tgtd.h"
+#include "target.h"
 #include "tgtadm_error.h"
 #include "util.h"
 #include "bs_thread.h"
 
 static LIST_HEAD(bst_list);
 
-static int sig_fd = -1;
-static LIST_HEAD(sig_finished_list);
-static pthread_mutex_t sig_finished_lock;
+struct bs_finish bs_finish, *bsf = &bs_finish;
+
+int sig_fd = -1;
+
+static int command_fd[2];
+static int done_fd[2];
+static pthread_t ack_thread;
+/* protected by pipe */
+static LIST_HEAD(ack_list);
+static pthread_cond_t finished_cond;
 
 int register_backingstore_template(struct backingstore_template *bst)
 {
@@ -66,12 +74,11 @@ struct backingstore_template *get_backingstore_template(const char *name)
 
 static void *bs_thread_ack_fn(void *arg)
 {
-	struct bs_thread_info *info = arg;
 	int command, ret, nr;
 	struct scsi_cmd *cmd;
 
 retry:
-	ret = read(info->command_fd[0], &command, sizeof(command));
+	ret = read(command_fd[0], &command, sizeof(command));
 	if (ret < 0) {
 		eprintf("ack pthread will be dead, %m\n");
 		if (errno == EAGAIN || errno == EINTR)
@@ -80,31 +87,28 @@ retry:
 		goto out;
 	}
 
-	if (info->stop)
-		goto out;
-
-	pthread_mutex_lock(&info->finished_lock);
+	pthread_mutex_lock(&bsf->finished_lock);
 retest:
-	if (list_empty(&info->finished_list)) {
-		pthread_cond_wait(&info->finished_cond, &info->finished_lock);
+	if (list_empty(&bsf->finished_list)) {
+		pthread_cond_wait(&finished_cond, &bsf->finished_lock);
 		goto retest;
 	}
 
-	while (!list_empty(&info->finished_list)) {
-		cmd = list_first_entry(&info->finished_list,
+	while (!list_empty(&bsf->finished_list)) {
+		cmd = list_first_entry(&bsf->finished_list,
 				 struct scsi_cmd, bs_list);
 
 		dprintf("found %p\n", cmd);
 
 		list_del(&cmd->bs_list);
-		list_add_tail(&cmd->bs_list, &info->ack_list);
+		list_add_tail(&cmd->bs_list, &ack_list);
 	}
 
-	pthread_mutex_unlock(&info->finished_lock);
+	pthread_mutex_unlock(&bsf->finished_lock);
 
 	nr = 1;
 rewrite:
-	ret = write(info->done_fd[1], &nr, sizeof(nr));
+	ret = write(done_fd[1], &nr, sizeof(nr));
 	if (ret < 0) {
 		eprintf("can't ack tgtd, %m\n");
 		if (errno == EAGAIN || errno == EINTR)
@@ -120,18 +124,17 @@ out:
 
 static void bs_thread_request_done(int fd, int events, void *data)
 {
-	struct bs_thread_info *info = data;
 	struct scsi_cmd *cmd;
 	int nr_events, ret;
 
-	ret = read(info->done_fd[0], &nr_events, sizeof(nr_events));
+	ret = read(done_fd[0], &nr_events, sizeof(nr_events));
 	if (ret < 0) {
 		eprintf("wrong wakeup\n");
 		return;
 	}
 
-	while (!list_empty(&info->ack_list)) {
-		cmd = list_first_entry(&info->ack_list,
+	while (!list_empty(&ack_list)) {
+		cmd = list_first_entry(&ack_list,
 				       struct scsi_cmd, bs_list);
 
 		dprintf("back to tgtd, %p\n", cmd);
@@ -141,7 +144,7 @@ static void bs_thread_request_done(int fd, int events, void *data)
 	}
 
 rewrite:
-	ret = write(info->command_fd[1], &nr_events, sizeof(nr_events));
+	ret = write(command_fd[1], &nr_events, sizeof(nr_events));
 	if (ret < 0) {
 		eprintf("can't write done, %m\n");
 		if (errno == EAGAIN || errno == EINTR)
@@ -154,6 +157,7 @@ rewrite:
 void bs_sig_request_done(int fd, int events, void *data)
 {
 	int ret;
+	struct bs_finish *b = data;
 	struct scsi_cmd *cmd;
 	struct signalfd_siginfo siginfo[16];
 	LIST_HEAD(list);
@@ -163,9 +167,9 @@ void bs_sig_request_done(int fd, int events, void *data)
 		return;
 	}
 
-	pthread_mutex_lock(&sig_finished_lock);
-	list_splice_init(&sig_finished_list, &list);
-	pthread_mutex_unlock(&sig_finished_lock);
+	pthread_mutex_lock(&b->finished_lock);
+	list_splice_init(&b->finished_list, &list);
+	pthread_mutex_unlock(&b->finished_lock);
 
 	while (!list_empty(&list)) {
 		cmd = list_first_entry(&list, struct scsi_cmd, bs_list);
@@ -181,6 +185,7 @@ static void *bs_thread_worker_fn(void *arg)
 	struct bs_thread_info *info = arg;
 	struct scsi_cmd *cmd;
 	sigset_t set;
+	struct bs_finish *tbsf;
 
 	sigfillset(&set);
 	sigprocmask(SIG_BLOCK, &set, NULL);
@@ -204,56 +209,126 @@ static void *bs_thread_worker_fn(void *arg)
 		cmd = list_first_entry(&info->pending_list,
 				       struct scsi_cmd, bs_list);
 
-		dprintf("got %p\n", cmd);
+
+		if (cmd->c_target->bsf)
+			tbsf = cmd->c_target->bsf;
+		else
+			tbsf = bsf;
 
 		list_del(&cmd->bs_list);
 		pthread_mutex_unlock(&info->pending_lock);
 
 		info->request_fn(cmd);
 
-		if (sig_fd < 0) {
-			pthread_mutex_lock(&info->finished_lock);
-			list_add_tail(&cmd->bs_list, &info->finished_list);
-			pthread_mutex_unlock(&info->finished_lock);
-
-			pthread_cond_signal(&info->finished_cond);
-		} else {
-			pthread_mutex_lock(&sig_finished_lock);
-			list_add_tail(&cmd->bs_list, &sig_finished_list);
-			pthread_mutex_unlock(&sig_finished_lock);
+		pthread_mutex_lock(&tbsf->finished_lock);
+		list_add_tail(&cmd->bs_list, &tbsf->finished_list);
+		pthread_mutex_unlock(&tbsf->finished_lock);
 
+		if (cmd->c_target->bsf)
+			pthread_kill(cmd->c_target->bsf->thread, SIGUSR2);
+		else if (sig_fd < 0)
+			pthread_cond_signal(&finished_cond);
+		else
 			kill(getpid(), SIGUSR2);
-		}
 	}
 
 	pthread_exit(NULL);
 }
 
-static void bs_init(void)
+static int bs_init_signalfd(struct bs_finish *b)
 {
-	static int done = 0;
 	sigset_t mask;
 	int ret;
 
-	if (done)
-		return;
-	done++;
-
 	sigemptyset(&mask);
 	sigaddset(&mask, SIGUSR2);
 	sigprocmask(SIG_BLOCK, &mask, NULL);
 
 	sig_fd = __signalfd(-1, &mask, 0);
 	if (sig_fd < 0)
-		return;
+		return 1;
 
-	ret = tgt_event_add(sig_fd, EPOLLIN, bs_sig_request_done, NULL);
+	ret = tgt_event_add(sig_fd, EPOLLIN, bs_sig_request_done, b);
 	if (ret < 0) {
 		close (sig_fd);
 		sig_fd = -1;
+
+		return 1;
 	}
 
-	pthread_mutex_init(&sig_finished_lock, NULL);
+	return 0;
+}
+
+static int bs_init_notify_thread(void)
+{
+	int ret;
+
+	pthread_cond_init(&finished_cond, NULL);
+
+	ret = pipe(command_fd);
+	if (ret) {
+		eprintf("failed to create command pipe, %m\n");
+		goto destroy_cond_mutex;
+	}
+
+	ret = pipe(done_fd);
+	if (ret) {
+		eprintf("failed to done command pipe, %m\n");
+		goto close_command_fd;
+	}
+
+	ret = tgt_event_add(done_fd[0], EPOLLIN, bs_thread_request_done, NULL);
+	if (ret) {
+		eprintf("failed to add epoll event\n");
+		goto close_done_fd;
+	}
+
+	ret = pthread_create(&ack_thread, NULL, bs_thread_ack_fn, NULL);
+	if (ret) {
+		eprintf("failed to create an ack thread, %s\n", strerror(ret));
+		goto event_del;
+	}
+
+	ret = write(command_fd[1], &ret, sizeof(ret));
+	if (ret <= 0)
+		goto event_del;
+
+	return 0;
+event_del:
+	tgt_event_del(done_fd[0]);
+
+close_done_fd:
+	close(done_fd[0]);
+	close(done_fd[1]);
+close_command_fd:
+	close(command_fd[0]);
+	close(command_fd[1]);
+destroy_cond_mutex:
+	pthread_cond_destroy(&finished_cond);
+
+	return 1;
+}
+
+int bs_init(void)
+{
+	int ret;
+
+	pthread_mutex_init(&bsf->finished_lock, NULL);
+	INIT_LIST_HEAD(&bsf->finished_list);
+
+	ret = bs_init_signalfd(bsf);
+	if (!ret) {
+		eprintf("use signalfd notification\n");
+		return 0;
+	}
+
+	ret = bs_init_notify_thread();
+	if (!ret) {
+		eprintf("use pthread notification\n");
+		return 0;
+	}
+
+	return 1;
 }
 
 int bs_thread_open(struct bs_thread_info *info, request_func_t *rfn,
@@ -261,48 +336,14 @@ int bs_thread_open(struct bs_thread_info *info, request_func_t *rfn,
 {
 	int i, ret;
 
-	bs_init();
-
 	info->request_fn = rfn;
 
-	INIT_LIST_HEAD(&info->ack_list);
-	INIT_LIST_HEAD(&info->finished_list);
 	INIT_LIST_HEAD(&info->pending_list);
 
-	pthread_cond_init(&info->finished_cond, NULL);
 	pthread_cond_init(&info->pending_cond, NULL);
-
-	pthread_mutex_init(&info->finished_lock, NULL);
 	pthread_mutex_init(&info->pending_lock, NULL);
 	pthread_mutex_init(&info->startup_lock, NULL);
 
-	ret = pipe(info->command_fd);
-	if (ret) {
-		eprintf("failed to create command pipe, %m\n");
-		goto destroy_cond_mutex;
-	}
-
-	ret = pipe(info->done_fd);
-	if (ret) {
-		eprintf("failed to done command pipe, %m\n");
-		goto close_command_fd;
-	}
-
-	if (sig_fd < 0) {
-		ret = tgt_event_add(info->done_fd[0], EPOLLIN, bs_thread_request_done,
-				    info);
-		if (ret) {
-			eprintf("failed to add epoll event\n");
-			goto close_done_fd;
-		}
-	}
-
-	ret = pthread_create(&info->ack_thread, NULL, bs_thread_ack_fn, info);
-	if (ret) {
-		eprintf("failed to create an ack thread, %s\n", strerror(ret));
-		goto event_del;
-	}
-
 	if (nr_threads > ARRAY_SIZE(info->worker_thread)) {
 		eprintf("too many threads %d\n", nr_threads);
 		nr_threads = ARRAY_SIZE(info->worker_thread);
@@ -321,41 +362,18 @@ int bs_thread_open(struct bs_thread_info *info, request_func_t *rfn,
 		}
 	}
 	pthread_mutex_unlock(&info->startup_lock);
-rewrite:
-	ret = write(info->command_fd[1], &ret, sizeof(ret));
-	if (ret < 0) {
-		eprintf("can't write done, %m\n");
-		if (errno == EAGAIN || errno == EINTR)
-			goto rewrite;
-	}
 
 	return 0;
 destroy_threads:
 	info->stop = 1;
-	write(info->command_fd[1], &ret, sizeof(ret));
-	pthread_join(info->ack_thread, NULL);
-
-	eprintf("stopped the ack thread\n");
 
 	pthread_mutex_unlock(&info->startup_lock);
 	for (; i > 0; i--) {
 		pthread_join(info->worker_thread[i - 1], NULL);
 		eprintf("stopped the worker thread %d\n", i - 1);
 	}
-event_del:
-	if (sig_fd < 0)
-		tgt_event_del(info->done_fd[0]);
 
-close_done_fd:
-	close(info->done_fd[0]);
-	close(info->done_fd[1]);
-close_command_fd:
-	close(info->command_fd[0]);
-	close(info->command_fd[1]);
-destroy_cond_mutex:
-	pthread_cond_destroy(&info->finished_cond);
 	pthread_cond_destroy(&info->pending_cond);
-	pthread_mutex_destroy(&info->finished_lock);
 	pthread_mutex_destroy(&info->pending_lock);
 	pthread_mutex_destroy(&info->startup_lock);
 
@@ -366,9 +384,6 @@ void bs_thread_close(struct bs_thread_info *info)
 {
 	int i;
 
-	pthread_cancel(info->ack_thread);
-	pthread_join(info->ack_thread, NULL);
-
 	info->stop = 1;
 	pthread_cond_broadcast(&info->pending_cond);
 
@@ -376,21 +391,10 @@ void bs_thread_close(struct bs_thread_info *info)
 		     i < ARRAY_SIZE(info->worker_thread); i++)
 		pthread_join(info->worker_thread[i], NULL);
 
-	pthread_cond_destroy(&info->finished_cond);
 	pthread_cond_destroy(&info->pending_cond);
-	pthread_mutex_destroy(&info->finished_lock);
 	pthread_mutex_destroy(&info->pending_lock);
 	pthread_mutex_destroy(&info->startup_lock);
 
-	if (sig_fd < 0)
-		tgt_event_del(info->done_fd[0]);
-
-	close(info->done_fd[0]);
-	close(info->done_fd[1]);
-
-	close(info->command_fd[0]);
-	close(info->command_fd[1]);
-
 	info->stop = 0;
 }
 
