@@ -224,8 +224,6 @@ static int waiting_rdma_slot;
 #define uint64_from_ptr(p) (uint64_t)(uintptr_t)(p)
 #define ptr_from_int64(p) (void *)(unsigned long)(p)
 
-#define ISCSI_LISTEN_PORT 3260
-
 /*
  * Crazy hard-coded linux iser settings need 128 * 8 slots + slop, plus
  * room for our rdmas and send requests.
@@ -1196,6 +1194,8 @@ static int iscsi_rdma_init(void)
 	INIT_LIST_HEAD(&iser_conn_list);
 	INIT_LIST_HEAD(&temp_conn);
 
+	iscsi_rdma_enabled = 1;
+
 	return ret;
 }
 
@@ -1248,6 +1248,11 @@ static int iscsi_rdma_login_complete(struct iscsi_connection *conn)
 	return ret;
 }
 
+static void iscsi_rdma_nexus_init(struct iscsi_connection *conn)
+{
+	conn->tp->ep_event_modify(conn, EPOLLIN);
+}
+
 /*
  * Copy the remote va and stag that were temporarily saved in conn_info.
  */
@@ -1727,6 +1732,7 @@ static struct iscsi_transport iscsi_iser = {
 	.data_padding		= 1,
 	.ep_init		= iscsi_rdma_init,
 	.ep_login_complete	= iscsi_rdma_login_complete,
+	.ep_nexus_init		= iscsi_rdma_nexus_init,
 	.alloc_task		= iscsi_iser_alloc_task,
 	.free_task		= iscsi_iser_free_task,
 	.ep_read		= iscsi_iser_read,
