@@ -38,6 +38,7 @@
 #include "util.h"
 #include "driver.h"
 #include "scsi.h"
+#include "tgtadm.h"
 #include "crc32c.h"
 
 #define MAX_QUEUE_CMD	128
@@ -73,6 +74,13 @@ enum {
 	IOSTATE_TX_END,
 };
 
+int iscsi_rdma_enabled;
+
+int iscsi_pthread_per_target(void)
+{
+	return sig_fd >= 0 && !iscsi_rdma_enabled;
+}
+
 void conn_read_pdu(struct iscsi_connection *conn)
 {
 	conn->rx_iostate = IOSTATE_RX_BHS;
@@ -467,6 +475,9 @@ static void login_start(struct iscsi_connection *conn)
 		}
 	}
 
+	if (conn->session_type == SESSION_DISCOVERY)
+		conn->tid = GLOBAL_TID;
+
 	if (conn->session_type == SESSION_NORMAL) {
 		if (!target_name) {
 			rsp->status_class = ISCSI_STATUS_CLS_INITIATOR_ERR;
@@ -482,8 +493,20 @@ static void login_start(struct iscsi_connection *conn)
 			conn->state = STATE_EXIT;
 			return;
 		}
+
 		conn->tid = target->tid;
 
+		if (target_redirected(target, conn)) {
+			char buf[NI_MAXHOST + NI_MAXSERV + 4];
+			snprintf(buf, sizeof(buf), "%s:%s", target->redirect_info.addr,
+				 target->redirect_info.port);
+			text_key_add(conn, "TargetAddress", buf);
+			rsp->status_class = ISCSI_STATUS_CLS_REDIRECT;
+			rsp->status_detail = target->redirect_info.reason;
+			conn->state = STATE_EXIT;
+			return;
+		}
+
 		if (tgt_get_target_state(target->tid) != SCSI_TARGET_READY) {
 			rsp->status_class = ISCSI_STATUS_CLS_TARGET_ERR;
 			rsp->status_detail = ISCSI_LOGIN_STATUS_TARGET_ERROR;
@@ -2224,7 +2247,7 @@ finish:
 		else {
 			conn->state = STATE_SCSI;
 			conn_read_pdu(conn);
-			conn->tp->ep_event_modify(conn, EPOLLIN);
+			conn->tp->ep_nexus_init(conn);
 		}
 		break;
 	case STATE_EXIT:
@@ -2317,8 +2340,14 @@ static int iscsi_param_parser(char *p)
 			else
 				len = strlen(addr);
 
-			iscsi_portal_addr = zalloc(len + 1);
-			memcpy(iscsi_portal_addr, addr, len);
+			if (iscsi_portal_addr) {
+				free(iscsi_portal_addr);
+				iscsi_portal_addr = NULL;
+			}
+			if (len) {
+				iscsi_portal_addr = zalloc(len + 1);
+				memcpy(iscsi_portal_addr, addr, len);
+			}
 		}
 
 		p += strcspn(p, ",");
