@@ -25,6 +25,7 @@
 #include <unistd.h>
 #include <netdb.h>
 #include <sys/stat.h>
+#include <sys/epoll.h>
 #include <sys/un.h>
 #include <netinet/in.h>
 #include <sys/socket.h>
@@ -32,10 +33,12 @@
 #include <netinet/tcp.h>
 #include <netinet/ip.h>
 #include <arpa/inet.h>
+#include <pthread.h>
 #include "iscsid.h"
 #include "tgtadm.h"
 #include "tgtd.h"
 #include "target.h"
+#include "util.h"
 
 LIST_HEAD(iscsi_targets_list);
 
@@ -189,6 +192,57 @@ int ip_acl(int tid, struct iscsi_connection *conn)
 	return -EPERM;
 }
 
+int target_redirected(struct iscsi_target *target, struct iscsi_connection *conn)
+{
+	struct sockaddr_storage from;
+	struct addrinfo hints, *res;
+	socklen_t len;
+	int ret;
+	char *p, *q, *str;
+
+	if (!strlen(target->redirect_info.addr))
+		return 0;
+
+	if (target->redirect_info.reason != ISCSI_LOGIN_STATUS_TGT_MOVED_TEMP &&
+	    target->redirect_info.reason != ISCSI_LOGIN_STATUS_TGT_MOVED_PERM)
+		return 0;
+
+	len = sizeof(from);
+	ret = conn->tp->ep_getpeername(conn, (struct sockaddr *)&from, &len);
+	if (ret < 0)
+		return 0;
+
+	p = strdup(target->redirect_info.addr);
+	if (!p)
+		return 0;
+	str = p;
+
+	if (*p == '[') {
+		p++;
+		if (!(q = strchr(p, ']'))) {
+			free(str);
+			return 0;
+		}
+		*(q++) = '\0';
+	}
+
+	memset(&hints, 0, sizeof(hints));
+	hints.ai_socktype = SOCK_STREAM;
+	hints.ai_flags = AI_NUMERICHOST;
+
+	ret = getaddrinfo(p, NULL, &hints, &res);
+	if (ret < 0) {
+		free(str);
+		return 0;
+	}
+
+	ret = address_match(res->ai_addr, (struct sockaddr *)&from);
+	freeaddrinfo(res);
+	free(str);
+
+	return !ret;
+}
+
 void target_list_build(struct iscsi_connection *conn, char *addr, char *name)
 {
 	struct iscsi_target *target;
@@ -252,12 +306,63 @@ void iscsi_target_destroy(int tid)
 	}
 
 	list_del(&target->tlist);
+
+	pthread_mutex_init(&target->event_lock, NULL);
+
+	if (target->bsfin.thread) {
+		target->stop_pthread = 1;
+		pthread_kill(target->bsfin.thread, SIGUSR2);
+
+		pthread_join(target->bsfin.thread, NULL);
+		pthread_mutex_destroy(&target->bsfin.finished_lock);
+	}
+
+	close(target->efd);
 	free(target);
 	isns_target_deregister(tgt_targetname(tid));
 
 	return;
 }
 
+static void *iscsi_thread_fn(void *arg)
+{
+	struct iscsi_target *t = arg;
+	struct epoll_event events[1024];
+	struct event_data *tev;
+	sigset_t mask;
+	int nevent, i;
+
+	sigemptyset(&mask);
+	sigaddset(&mask, SIGUSR2);
+	pthread_sigmask(SIG_BLOCK, &mask, NULL);
+
+	pthread_mutex_lock(&t->event_lock);
+
+	do_tgt_event_add(t->efd, &t->events_list, sig_fd, EPOLLIN,
+			 bs_sig_request_done, &t->bsfin);
+
+	pthread_mutex_unlock(&t->event_lock);
+
+retry:
+	nevent = epoll_wait(t->efd, events, ARRAY_SIZE(events), 1000);
+	if (nevent < 0) {
+		if (errno != EINTR) {
+			eprintf("%m\n");
+			exit(1);
+		}
+	} else if (nevent) {
+		for (i = 0; i < nevent; i++) {
+			tev = (struct event_data *) events[i].data.ptr;
+			tev->handler(tev->fd, events[i].events, tev->data);
+		}
+	}
+
+	if (!t->stop_pthread)
+		goto retry;
+
+	pthread_exit(NULL);
+}
+
 int iscsi_target_create(struct target *t)
 {
 	int tid = t->tid;
@@ -288,11 +393,15 @@ int iscsi_target_create(struct target *t)
 		[ISCSI_PARAM_MAX_OUTST_PDU] =  {0, 0},  /* not in open-iscsi */
 	};
 
-	target = malloc(sizeof(*target));
+	target = zalloc(sizeof(*target));
 	if (!target)
 		return -ENOMEM;
 
-	memset(target, 0, sizeof(*target));
+	target->efd = epoll_create(128);
+	if (target->efd < 0) {
+		free(target);
+		return -EINVAL;
+	}
 
 	memcpy(target->session_param, default_tgt_session_param,
 	       sizeof(target->session_param));
@@ -300,10 +409,21 @@ int iscsi_target_create(struct target *t)
 	INIT_LIST_HEAD(&target->tlist);
 	INIT_LIST_HEAD(&target->sessions_list);
 	INIT_LIST_HEAD(&target->isns_list);
+	INIT_LIST_HEAD(&target->events_list);
 	target->tid = tid;
 	list_add_tail(&target->tlist, &iscsi_targets_list);
 
 	isns_target_register(tgt_targetname(tid));
+
+	if (iscsi_pthread_per_target()) {
+		pthread_create(&target->bsfin.thread, NULL, iscsi_thread_fn, target);
+
+		pthread_mutex_init(&target->bsfin.finished_lock, NULL);
+		INIT_LIST_HEAD(&target->bsfin.finished_list);
+		t->bsf = &target->bsfin;
+		eprintf("create thread %u\n", (unsigned)target->bsfin.thread);
+	}
+
 	return 0;
 }
 
@@ -347,6 +467,29 @@ int iscsi_target_update(int mode, int op, int tid, uint64_t sid, uint64_t lun,
 
 		dprintf("%s:%s\n", name, str);
 
+		if (!strncmp(name, "RedirectAddress", 15)) {
+			snprintf(target->redirect_info.addr,
+				 sizeof(target->redirect_info.addr), "%s", str);
+			err = TGTADM_SUCCESS;
+			break;
+		} else if (!strncmp(name, "RedirectPort", 12)) {
+			snprintf(target->redirect_info.port,
+				 sizeof(target->redirect_info.port), "%s", str);
+			err = TGTADM_SUCCESS;
+			break;
+		} else if (!strncmp(name, "RedirectReason", 14)) {
+			if (!strncmp(str, "Temporary", 9)) {
+				target->redirect_info.reason =
+					ISCSI_LOGIN_STATUS_TGT_MOVED_TEMP;
+				err = TGTADM_SUCCESS;
+			} else if (!strncmp(str, "Permanent", 9)) {
+				target->redirect_info.reason =
+					ISCSI_LOGIN_STATUS_TGT_MOVED_PERM;
+				err = TGTADM_SUCCESS;
+			} else
+				break;
+		}
+
 		idx = param_index_by_name(name, session_keys);
 		if (idx >= 0) {
 			err = iscsi_session_param_update(target, idx, str);
@@ -394,6 +537,37 @@ static int iscsi_target_show_session(struct iscsi_target* target, uint64_t sid,
 	return total;
 }
 
+#define __buffer_check(buf, total, len, rest)	\
+({						\
+	buf += len;				\
+	total += len;				\
+	rest -= len;				\
+	if (!rest)				\
+		return total;			\
+})
+
+static int show_redirect_info(struct iscsi_target* target, char *buf, int rest)
+{
+	int len, total = 0;
+
+	len = snprintf(buf, rest, "RedirectAddress=%s\n", target->redirect_info.addr);
+	__buffer_check(buf, total, len, rest);
+	len = snprintf(buf, rest, "RedirectPort=%s\n", target->redirect_info.port);
+	__buffer_check(buf, total, len, rest);
+	if (target->redirect_info.reason == ISCSI_LOGIN_STATUS_TGT_MOVED_TEMP) {
+		len = snprintf(buf, rest, "RedirectReason=Temporary\n");
+		__buffer_check(buf, total, len, rest);
+	} else if (target->redirect_info.reason == ISCSI_LOGIN_STATUS_TGT_MOVED_PERM) {
+		len = snprintf(buf, rest, "RedirectReason=Permanent\n");
+		__buffer_check(buf, total, len, rest);
+	} else {
+		len = snprintf(buf, rest, "RedirectReason=Unknown\n");
+		__buffer_check(buf, total, len, rest);
+	}
+
+	return total;
+}
+
 int iscsi_target_show(int mode, int tid, uint64_t sid, uint32_t cid, uint64_t lun,
 		      char *buf, int rest)
 {
@@ -411,7 +585,10 @@ int iscsi_target_show(int mode, int tid, uint64_t sid, uint32_t cid, uint64_t lu
 		total = isns_show(buf, rest);
 		break;
 	case MODE_TARGET:
-		len = show_iscsi_param(buf, target->session_param, rest);
+		if (strlen(target->redirect_info.addr))
+			len = show_redirect_info(target, buf, rest);
+		else
+			len = show_iscsi_param(buf, target->session_param, rest);
 		total += len;
 		break;
 	case MODE_SESSION:
