@@ -37,7 +37,6 @@
 #include "spc.h"
 
 #define PRODUCT_REV	"0"
-#define BLK_SHIFT	9
 
 /*
  * Protocol Identifier Values
@@ -112,7 +111,7 @@
 
 static void update_vpd_80(struct scsi_lu *lu, void *sn)
 {
-	struct vpd *vpd_pg = lu->attrs.lu_vpd[0];
+	struct vpd *vpd_pg = lu->attrs.lu_vpd[PCODE_OFFSET(0x80)];
 	char *data = (char *)vpd_pg->data;
 
 	memset(data, 0x20, vpd_pg->size);
@@ -123,14 +122,14 @@ static void update_vpd_80(struct scsi_lu *lu, void *sn)
 
 		p = data + vpd_pg->size - 1;
 		q = sn + tmp - 1;
-		for (; tmp > 0; tmp--, q)
+		for (; tmp > 0; tmp--)
 			*(p--) = *(q--);
 	}
 }
 
 static void update_vpd_83(struct scsi_lu *lu, void *id)
 {
-	struct vpd *vpd_pg = lu->attrs.lu_vpd[3];
+	struct vpd *vpd_pg = lu->attrs.lu_vpd[PCODE_OFFSET(0x83)];
 	uint8_t	*data = vpd_pg->data;
 
 	data[0] = INQ_CODE_ASCII;
