@@ -358,7 +358,7 @@ static int str_to_device_type(char *str)
 	else if (!strcmp(str, "ssc"))
 		return TYPE_TAPE;
 	else if (!strcmp(str, "pt"))
-		return TYPE_SPT;
+		return TYPE_PT;
 	else {
 		eprintf("unknown target type: %s\n", str);
 		exit(EINVAL);
@@ -554,6 +554,31 @@ int main(int argc, char **argv)
 		exit(EINVAL);
 	}
 
+	if (mode == MODE_SYSTEM) {
+		switch (op) {
+		case OP_UPDATE:
+			rc = verify_mode_params(argc, argv, "LmonvC");
+			if (rc) {
+				eprintf("system mode: option '-%c' is not "
+					"allowed/supported\n", rc);
+				exit(EINVAL);
+			}
+			if ((!name || !value)) {
+				eprintf("update operation requires 'name'"
+					" and 'value' options\n");
+				exit(EINVAL);
+			}
+			break;
+		case OP_SHOW:
+		case OP_DELETE:
+			break;
+		default:
+			eprintf("option %d not supported in system mode\n", op);
+			exit(EINVAL);
+			break;
+		}
+	}
+
 	if (mode == MODE_TARGET) {
 		if ((tid <= 0 && (op != OP_SHOW))) {
 			eprintf("'tid' option is necessary\n");
@@ -657,10 +682,8 @@ int main(int argc, char **argv)
 				eprintf("'user' option is necessary\n");
 				exit(EINVAL);
 			}
-			if (tid <= 0) {
-				eprintf("'tid' option is necessary\n");
-				exit(EINVAL);
-			}
+			if (tid == -1)
+				tid = GLOBAL_TID;
 			break;
 		case OP_UNBIND:
 			rc = verify_mode_params(argc, argv, "LmouC");
@@ -673,10 +696,8 @@ int main(int argc, char **argv)
 				eprintf("'user' option is necessary\n");
 				exit(EINVAL);
 			}
-			if (tid <= 0) {
-				eprintf("'tid' option is necessary\n");
-				exit(EINVAL);
-			}
+			if (tid == -1)
+				tid = GLOBAL_TID;
 			break;
 		default:
 			eprintf("option %d not supported in account mode\n", op);
