@@ -135,7 +135,8 @@ set_rlimit:
 	return 0;
 }
 
-int tgt_event_add(int fd, int events, event_handler_t handler, void *data)
+int do_tgt_event_add(int efd, struct list_head *list, int fd, int events,
+		     event_handler_t handler, void *data)
 {
 	struct epoll_event ev;
 	struct event_data *tev;
@@ -152,39 +153,45 @@ int tgt_event_add(int fd, int events, event_handler_t handler, void *data)
 	memset(&ev, 0, sizeof(ev));
 	ev.events = events;
 	ev.data.ptr = tev;
-	err = epoll_ctl(ep_fd, EPOLL_CTL_ADD, fd, &ev);
+	err = epoll_ctl(efd, EPOLL_CTL_ADD, fd, &ev);
 	if (err) {
 		eprintf("Cannot add fd, %m\n");
 		free(tev);
 	} else
-		list_add(&tev->e_list, &tgt_events_list);
+		list_add(&tev->e_list, list);
 
 	return err;
 }
 
-static struct event_data *tgt_event_lookup(int fd)
+int tgt_event_add(int fd, int events, event_handler_t handler, void *data)
+{
+	return do_tgt_event_add(ep_fd, &tgt_events_list, fd, events, handler,
+				data);
+}
+
+static struct event_data *tgt_event_lookup(struct list_head *list, int fd)
 {
 	struct event_data *tev;
 
-	list_for_each_entry(tev, &tgt_events_list, e_list) {
+	list_for_each_entry(tev, list, e_list) {
 		if (tev->fd == fd)
 			return tev;
 	}
 	return NULL;
 }
 
-void tgt_event_del(int fd)
+void do_tgt_event_del(int efd, struct list_head *list, int fd)
 {
 	struct event_data *tev;
 	int ret;
 
-	tev = tgt_event_lookup(fd);
+	tev = tgt_event_lookup(list, fd);
 	if (!tev) {
 		eprintf("Cannot find event %d\n", fd);
 		return;
 	}
 
-	ret = epoll_ctl(ep_fd, EPOLL_CTL_DEL, fd, NULL);
+	ret = epoll_ctl(efd, EPOLL_CTL_DEL, fd, NULL);
 	if (ret < 0)
 		eprintf("fail to remove epoll event, %s\n", strerror(errno));
 
@@ -192,12 +199,17 @@ void tgt_event_del(int fd)
 	free(tev);
 }
 
-int tgt_event_modify(int fd, int events)
+void tgt_event_del(int fd)
+{
+	do_tgt_event_del(ep_fd, &tgt_events_list, fd);
+}
+
+int do_tgt_event_modify(int efd, struct list_head *list, int fd, int events)
 {
 	struct epoll_event ev;
 	struct event_data *tev;
 
-	tev = tgt_event_lookup(fd);
+	tev = tgt_event_lookup(list, fd);
 	if (!tev) {
 		eprintf("Cannot find event %d\n", fd);
 		return -EINVAL;
@@ -207,7 +219,12 @@ int tgt_event_modify(int fd, int events)
 	ev.events = events;
 	ev.data.ptr = tev;
 
-	return epoll_ctl(ep_fd, EPOLL_CTL_MOD, fd, &ev);
+	return epoll_ctl(efd, EPOLL_CTL_MOD, fd, &ev);
+}
+
+int tgt_event_modify(int fd, int events)
+{
+	return do_tgt_event_modify(ep_fd, &tgt_events_list, fd, events);
 }
 
 void tgt_init_sched_event(struct event_data *evt,
@@ -235,6 +252,69 @@ void tgt_remove_sched_event(struct event_data *evt)
 	}
 }
 
+struct ext_prog_info {
+	void (*callback)(void *data, int result);
+	void *data;
+};
+
+static void run_ext_callback(int fd, int events, void *data)
+{
+	int ret, result;
+	struct ext_prog_info *ex = data;
+
+	ret = read(fd, &result, sizeof(result));
+	if (ret != sizeof(result)) {
+		result = -EINVAL;
+		eprintf("failed to get the result.");
+	}
+
+	if (ex->callback)
+		ex->callback(ex->data, result);
+
+	tgt_event_del(fd);
+	close(fd);
+	free(data);
+}
+
+int run_ext_program(const char *cmd,
+		    void (*callback)(void *data, int result), void *data)
+{
+	pid_t pid;
+	int fds[2], ret;
+	struct ext_prog_info *ex;
+
+	ex = zalloc(sizeof(*ex));
+	if (!ex)
+		return -ENOMEM;
+
+	ret = pipe(fds);
+	if (ret < 0) {
+		free(ex);
+		return ret;
+	}
+
+	eprintf("%d %d\n", fds[0], fds[1]);
+
+	ex->callback = callback;
+	ex->data = data;
+
+	tgt_event_add(fds[0], EPOLLIN, run_ext_callback, ex);
+
+	pid = fork();
+	if (pid < 0)
+		return pid;
+
+	if (!pid) {
+		ret = system(cmd);
+		write(fds[1], &ret, sizeof(ret));
+		return 0;
+	}
+
+	close(fds[1]);
+
+	return 0;
+}
+
 static int tgt_exec_scheduled(void)
 {
 	struct list_head *last_sched;
@@ -445,6 +525,8 @@ int main(int argc, char **argv)
 		}
 	}
 
+	bs_init();
+
 	event_loop();
 
 	lld_exit();
